#(c) 2016-2018 by Authors
#This file is a part of Flye program.
#Released under the BSD license (see LICENSE file)

"""
File with Trestle configurations
"""

vals = {
        "max_iter" : 10,
        "buffer_count" : 3,
        "min_edge_cov" : 10,
        "min_aln_rate" : 0.5,
        "min_bridge_count" : 5,
        "min_bridge_factor" : 2,
        "min_mult" : 2,
        "max_mult" : 3,
        "flanking_len" : 10000,
        "sub_thresh" : 0.1,
        "del_thresh" : 0.2,
        "ins_thresh" : 0.3,
        "cons_aln_rate" : 0.01,
        "min_supp_align_len" : 1000,
        "max_supp_align_overlap" : 100,
        "orientations_to_run" : "forward",
        "num_pol_iters" : 1
    }
